/****************************************************************************

    This file is part of The Newtonator soft synth.

    Copyright (C) 2010-2011 Michael Bechard <mbechard@users.sourceforge.net>

    The Newtonator is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The Newtonator is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with The Newtonator.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************************/

#ifndef NEWTONATOR2_H_INCLUDED
#define NEWTONATOR2_H_INCLUDED

#include "INewtSettings.h"
#include "CachedPrecomputedPow.h"
#include <queue>
#include <stk/Delay.h>

class Newtonator2
{
    public:
    Newtonator2(unsigned int channels = 2);
    virtual ~Newtonator2();
    inline VariSource* getGravModSources() { return _gravModSource; };
    inline stk::ADSR* getAmpEnvs() { return _ampEnv; };
    stk::StkFloat newtTick(unsigned int channel, unsigned int frameIdx);
    virtual void initChannels(unsigned int numChannels);
    void setSettings(INewtSettings* set);
    static CeilingBehavEnum GetCBFromFloat(stk::StkFloat val);


    //Note control
    bool isPlaying() const;
    void keyOn(stk::StkFloat velocity);
    void keyOff(stk::StkFloat velocity);
    void setFrequency(stk::StkFloat frequency);
    void bendFrequency(stk::StkFloat factor);
    inline stk::StkFloat getFrequency() const { return _actualFrequency; }
    inline bool needsKeyOff() const { return _needsKeyOff; }

    //Accessors for user control variables
    inline stk::StkFloat getGravModDepth(unsigned int chan) const { return separate(_set->getGModDepth(chan), chan); };
    inline stk::StkFloat getGravityScale(unsigned int chan) const { return _set->getGravityScale(chan); };
    inline CeilingBehavEnum getCeilingBehavior(unsigned int chan) const { return _set->getCeilingBehavior(chan); };
    inline stk::StkFloat getCeilingBounce(unsigned int chan) const { return (_set->getCeilingBounce(chan) * -1) - 1.00001; }
    inline stk::StkFloat getVelocReset(unsigned int chan) const { return _set->getVelocReset(chan); }
    inline bool getStereo() const { return _set->getStereo(); }
    inline int numChannels() const { return (_set->getStereo() ? 2 : 1); }
    inline stk::StkFloat getVelocDelay(unsigned int chan) { return _set->getVelocDelay(chan); }
    inline stk::StkFloat getDeadZone(unsigned int chan) { return _set->getDeadZone(chan); }
    inline stk::StkFloat getWobbleA(unsigned int chan) { return _set->getWobbleA(chan); }
    inline stk::StkFloat getWobbleB(unsigned int chan) { return _set->getWobbleB(chan); }
    inline stk::StkFloat getWobbleScale(unsigned int chan) { return _set->getWobbleScale(chan); }
    inline stk::StkFloat getWobbleTimbre(unsigned int chan) { return _set->getWobbleTimbre(chan); }
    inline stk::StkFloat getGModDCOffset(unsigned int chan) { return _set->getGModDCOffset(chan); }
    inline stk::StkFloat getVelocRM(unsigned int chan) { return _set->getVelocRM(chan); }
    inline stk::StkFloat getVelocSaH(unsigned int chan) { return _set->getVelocSaH(chan) * (440 / _noteFrequency); }
    inline stk::StkFloat getAmp(unsigned int chan) { return _cpMainAmp.Compute(_set->getAmp(chan), 2); }
    inline stk::StkFloat getChanSep(unsigned int chan) { return .25; }

    inline stk::StkFloat getCeilingTop(unsigned int chan) { return MAX_VAL; }
    inline stk::StkFloat getCeilingBottom(unsigned int chan) { return MIN_VAL; }

    private:
    stk::StkFloat* _pPrevPos;
    stk::StkFloat* _pPrevVeloc;
    stk::StkFloat* _pHoldVeloc;
    stk::StkFloat _gravity;
    stk::StkFloat _noteFrequency;
    stk::StkFloat _actualFrequency;
    stk::StkFloat _timeForCycle;
    stk::StkFloat _noteAmp;
    stk::StkFloat _noteCurrMaxRM;
    unsigned int _currVelocSaH;
    int* _gravDir;
    VariSource* _gravModSource;
    stk::ADSR* _ampEnv;
    stk::Delay* _velocDelay;
    unsigned int _numChannels;
    stk::StkFloat _sampCount;
    const stk::StkFloat MAX_VAL;
    const stk::StkFloat MIN_VAL;
	bool _needsKeyOff;

    //User Control values
	INewtSettings* _set;

    //Clip Wobble settings
    bool* _isClipping;
    stk::SineWave* _wobbleWav;
    stk::StkFloat* _a;
    void resetWobble(unsigned int chan);

    //Cached computed values
    CachedPrecomputedPow _cpGModFreqMult;
    CachedPrecomputedPow _cpWobbleFreq;
    CachedPrecomputedPow _cpMainAmp;

    //Methods
    void clearChannelMemory();
    void resetForFreqChange();
	void syncSynthSettings();
	void velocResetForSampAndHold(unsigned int channel, stk::StkFloat velocReset);
	stk::StkFloat separate(stk::StkFloat val, unsigned int chan) const;
	stk::StkFloat procVeloc(stk::StkFloat veloc, unsigned int chan);
	stk::StkFloat wobbleClip(stk::StkFloat currSample, stk::StkFloat modifier, unsigned int chan);
	stk::StkFloat ringMod(stk::StkFloat carrier, stk::StkFloat mod, unsigned int chan);
	stk::StkFloat doVelocDelay(stk::StkFloat veloc, unsigned int chan);
	stk::StkFloat doVelocSaH(stk::StkFloat veloc, unsigned int chan);
    stk::StkFloat limitVal(stk::StkFloat val, stk::StkFloat limit);
};


#endif // NEWTONATOR2_H_INCLUDED
